import numpy as np
import torch
import torch.nn as nn

import sys
from timeit import default_timer as timer

import dataset_util as utils

class Model_CUDA:
    def __init__(self, net, optimizer, criterion, training_set, validation_set, batch_size_train=128, batch_size_val=None):
        self._net = net
        self._opt = optimizer
        self._criterion = criterion
        self._train_set, self._val_set = training_set, validation_set
        if batch_size_val == None:
            batch_size_val = 2 * batch_size_train
        # set training set and test set data loaders from utils
        self._train_loader = utils.random_batch_loader(self._train_set, batch_size_train)
        self._val_loader = utils.batch_loader(self._val_set, batch_size_val)
        self._epo = 0
    
    def val_loss(self):
        sum_loss = torch.tensor(0.).cuda()
        batch_count = 0
        for batch in self._val_loader:
            coord = batch.coords.cuda().requires_grad_()
            force = batch.forces.cuda()
            pred_force = self._net.forward(coord)
            batch_loss = self._criterion(pred_force, force)
            sum_loss += batch_loss.data
            batch_count += 1
        return sum_loss.item() / batch_count
    
    def train_epoch(self, verbose=False):
        o = sys.stdout
        start = timer()
        batch_count = 0
        sum_loss = torch.tensor(0.).cuda()
        for num, batch in enumerate(self._train_loader):
            self._opt.zero_grad()
            coord = batch.coords.cuda().requires_grad_()
            force = batch.forces.cuda()
            pred_force = self._net.forward(coord)
            batch_loss = self._criterion(pred_force, force)
            sum_loss += batch_loss.data
            batch_loss.backward()
            self._opt.step()
            #break
            if verbose and num % 200 == 199:
                o.write("\rBatch: {}, training loss: {:.2f}.".format(num + 1, sum_loss / (num + 1)))
                o.flush()
            batch_count += 1
        if verbose:
            o.write("\n")
            o.flush()
        stop = timer()
        self._epo += 1
        val_loss_value = self.val_loss()
        if verbose:
            o.write("Epoch %d, time consumed for this epoch: %.2fs, Val loss: %.2f\n" % (self._epo, (stop - start), val_loss_value))
        return {"epoch":self._epo, "time":(stop - start), "val_loss":val_loss_value, "train_loss": sum_loss.item() / batch_count}

